import json
import logging
import os
import shutil

import pytest

# Formatting of testing log messages.
from dls_logformatter.dls_logformatter import DlsLogformatter

# Version of the package.
from dls_mainiac_lib.version import meta as version_meta

logger = logging.getLogger(__name__)


# --------------------------------------------------------------------------------
@pytest.fixture(scope="function")
def constants(request):

    constants = {}

    constants["mainiac_database_endpoint"] = "%s/mainiac.sqlite"
    constants["mainiac_port"] = 15006
    constants["mainiac_server_endpoint"] = "tcp://*:%d" % (constants["mainiac_port"])
    constants["mainiac_client_endpoint"] = "tcp://localhost:%d" % (
        constants["mainiac_port"]
    )

    constants["protocolj_port"] = 15007
    constants["protocolj_client_endpoint"] = "http://localhost:%d" % (
        constants["protocolj_port"]
    )

    yield constants


# --------------------------------------------------------------------------------
@pytest.fixture()
def logging_setup():

    formatter = DlsLogformatter(type="long")
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logging.getLogger().addHandler(handler)

    # Log level for all modules.
    logging.getLogger().setLevel("DEBUG")

    # Cover the version.
    logger.info("\n%s", (json.dumps(version_meta(), indent=4)))

    yield None


# --------------------------------------------------------------------------------
@pytest.fixture(scope="function")
def output_directory(request):

    # Tmp directory which we can write into.
    output_directory = "/tmp/%s/%s/%s" % (
        "/".join(__file__.split("/")[-3:-1]),
        request.cls.__name__,
        request.function.__name__,
    )

    # Tmp directory which we can write into.
    if os.path.exists(output_directory):
        shutil.rmtree(output_directory, ignore_errors=False, onerror=None)
    os.makedirs(output_directory)

    # logger.debug("output_directory is %s" % (output_directory))

    yield output_directory



import signal
import sys
import time

from dls_utilpack.signal import Signal


def main():
    """
    Wait a bit then die unless interrupt comes first, exit code says what happened.
    """

    try:
        sleep_time = 2.0

        # Catch sigint signals and just count them.
        sigint_signal = Signal(signal.SIGINT)
        sigint_signal.activate()

        end_time = time.time() + sleep_time
        while time.time() < end_time:
            if sigint_signal.count() > 0:
                sys.exit(2)

        sys.exit(1)
    except Exception:
        sys.exit(-1)


# ---------------------------------------------------------------------
# Run the main program.
main()



import argparse
import logging
import os
import shutil

import pytest

# Class under test.
from dls_mainiac_lib.mainiac import Mainiac

logger = logging.getLogger(__name__)


class Test_04_rotating:

    # ----------------------------------------------------------------------------------------
    def test(
        self,
        constants,
        logging_setup,
        output_directory,
        capsys,
    ):
        """
        Test mainiac base class.
        """

        failure_message = None
        try:

            # Instantiate the app class.
            app = _App()

            # Configure the app from command line arguments.
            app.parse_args_and_configure_logging([])

            # Run the gui wrapped in a try/catch.
            app.try_run_catch()

        except Exception as exception:
            logger.exception("unexpected exception during the test", exc_info=exception)
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)


# ---------------------------------------------------------------------------------
class _App(Mainiac):
    """
    App class.
    """

    def __init__(
        self,
    ):

        program_name = "test_04_rotating"

        self.__logfile_directory = f"/tmp/logs/{program_name}"

        if os.path.isdir(self.__logfile_directory):
            shutil.rmtree(self.__logfile_directory)

        Mainiac.__init__(self, program_name)

    # ----------------------------------------------------------
    def run(self):
        pad = "-" * 100

        for i in range(0, 6):
            logger.info(f"info message{i} {pad}")

        for i in range(0, 5):
            if i == 0:
                filename = "logform.log"
            else:
                filename = f"logform.log.{i}"

            filename = f"{self.__logfile_directory}/{filename}"

            # Check the logfile got written.
            assert os.path.exists(filename)

            # Check the message is the correct rotation.
            message = f"message{5-i}"
            with open(filename, "r") as stream:
                lines = stream.readlines()
                assert len(lines) == 1
                assert message in lines[0]

        assert not os.path.exists(f"{self.__logfile_directory}/logform.5")

    # --------------------------------------------------------------------------
    def configure_logging(self, settings=None):
        """
        Configure runtime logging, override base class.
        Presume that self._args is already set.
        """

        settings = {
            "logfile": {
                "enabled": True,
                "max_bytes": 100,
                "backup_count": 4,
            }
        }
        # Call the base method which has the extra kwarg.
        Mainiac.configure_logging(self, settings)

    # ----------------------------------------------------------
    def version(self):
        return "x.y.z"

    # ----------------------------------------------------------
    def about(self):
        return {"url": "/some/good/url", "description": "A good description"}

    # ----------------------------------------------------------
    def build_parser(self, arglist=None):

        # Make a parser.
        parser = argparse.ArgumentParser()

        # Return a structure describing the default subcommand.
        return parser



import contextlib
import io
import logging
import sys

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def captured_output():
    new_out, new_err = io.StringIO(), io.StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


# ----------------------------------------------------------------------------------------
def assert_parse_system_exit(
    case,
    app,
    arglist,
    exit_code=None,
    stdout_contains=None,
    stderr_contains=None,
):
    """
    Make sure the expected parse exit is handled.
    """

    # Capture the printing of the help and syntax messages.
    with captured_output() as (captured_stdout, captured_stderr):
        # with pytest.raises(SystemExit) as exc_info:
        exception = None
        try:
            app.parse_args(arglist)
        except SystemExit as e:
            exception = e

        if exception is None:
            assert False, "%s did not raise SystemExit" % (case)

        # TODO: Figure out how to assert exit code from SystemExit exception.
        # assert exc_info.value.code == exit_code

        captured_stdout.seek(0)
        captured_stdout = captured_stdout.read()

        captured_stderr.seek(0)
        captured_stderr = captured_stderr.read()

        if stdout_contains is not None:
            if stdout_contains not in captured_stdout:
                t = (
                    "failed %s %s"
                    "\n stdout expected to contain: %s"
                    "\n but is actually:\n%s"
                )
                logger.info(
                    t % (case, str(arglist), stdout_contains, captured_stdout),
                )
                assert False, case
        else:
            if captured_stdout != "":
                logger.info("stdout: %s" % (captured_stdout))

        if stderr_contains is not None:
            if stderr_contains not in captured_stderr:
                t = (
                    "failed %s %s"
                    "\nstderr expected to contain: %s"
                    "\n but is actually:\n%s"
                )
                logger.info(
                    t % (case, str(arglist), stderr_contains, captured_stderr),
                )
                assert False, case
        else:
            if captured_stderr != "":
                logger.info("stderr: %s" % (captured_stderr))


# ----------------------------------------------------------------------------------------
def assert_parse_success(case, app, arglist):

    # Capture the printing of the help and syntax messages.
    with captured_output() as (captured_stdout, captured_stderr):
        try:
            app.parse_args(arglist)
        except Exception:
            captured_stderr.seek(0)
            captured_stderr = captured_stderr.read()
            logger.info(
                "failed %s %s\nstderr:\n%s" % (case, str(arglist), captured_stderr)
            )
            assert False, case



import logging
import os
import time
from subprocess import Popen

import psutil
import pytest

# Function under test.
from dls_mainiac_lib.duplicate import interrupt_duplicate_commandlines

logger = logging.getLogger(__name__)


class Test_05_duplicate:

    # ----------------------------------------------------------------------------------------
    def test_05_duplicate(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """
        Test discovery and kill of duplicate commandline processes.
        """

        failure_message = None
        try:

            processes = []

            # Fire up three identical processes.
            for i in range(3):
                args = []

                args.append("python3")
                args.append("-m")
                args.append("tests.wilt")
                args.append(output_directory)

                # Run the process.
                pid = Popen(
                    args,
                    shell=False,
                    stdin=None,
                    stdout=None,
                    stderr=None,
                    close_fds=True,
                    preexec_fn=os.setsid,
                ).pid

                processes.append(psutil.Process(pid))

            # Give the processes a chance to become alive.
            time.sleep(0.5)

            # Kill all duplicates of the first.
            interrupt_duplicate_commandlines(this_process=processes[0])

            # Wait for all processes to die.
            gone, alive = psutil.wait_procs(processes, timeout=2.5)
            assert len(gone) == 3

            # First one dies of old age, the other two die of interrupt.
            assert processes[0].returncode == 1
            assert processes[1].returncode == 2
            assert processes[2].returncode == 2

        except Exception as exception:
            logger.exception(
                "unexpected exception during the test",
                exc_info=exception,
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)



import argparse
import logging
import multiprocessing
import os
import shutil
import time

import pytest

# Class under test.
from dls_mainiac_lib.mainiac import Mainiac

logger = logging.getLogger(__name__)


class Test_06_mpqueue:

    # ----------------------------------------------------------------------------------------
    def test(
        self,
        constants,
        logging_setup,
        output_directory,
        capsys,
    ):
        """
        Test mainiac base class.
        """

        failure_message = None
        try:

            # Instantiate the app class.
            app = _App()

            # logging.getLogger().handlers = []
            # for handler in logging.getLogger().handlers:
            #    print(f"********** handler is a {type(handler).__name__}")

            # Configure the app from command line arguments.
            app.parse_args_and_configure_logging([])

            # Call the run method of the object, wrapped in a try/catch.
            app.try_run_catch()

        except Exception as exception:
            logger.exception("unexpected exception during the test", exc_info=exception)
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)


# ---------------------------------------------------------------------------------
class _App(Mainiac):
    """
    App class.
    """

    def __init__(
        self,
    ):

        program_name = "test_06_mpqueue"

        self.__logfile_directory = f"/tmp/logs/{program_name}"

        if os.path.isdir(self.__logfile_directory):
            shutil.rmtree(self.__logfile_directory)

        Mainiac.__init__(self, program_name)

        self.__count = 5

    # ----------------------------------------------------------
    def _run_in_process(self, mpqueue):
        multiprocessing.current_process().name = "process"

        # Remove existing handlers.
        logging.getLogger().handlers = []
        mpqueue_handler = logging.handlers.QueueHandler(mpqueue)
        mpqueue_handler.setLevel(logging.DEBUG)
        logging.getLogger().addHandler(mpqueue_handler)

        pad = "-" * 100

        for i in range(0, self.__count):
            logger.info(f"info message{i} {pad}")

    # ----------------------------------------------------------
    def run(self):
        logger.info(f"master pid is {os.getpid()}")
        process = multiprocessing.Process(
            target=self._run_in_process, args=[self.mpqueue]
        )
        process.start()
        process.join()

        time.sleep(0.2)

        # Verify we heard all the log entries in the mpqueue.
        assert self.__count == self.mpqueue_heard_count, "mpqueue_heard_count"

        for i in range(0, self.__count):
            if i == 0:
                filename = "logform.log"
            else:
                filename = f"logform.log.{i}"

            filename = f"{self.__logfile_directory}/{filename}"

            # Check the logfile got written.
            assert os.path.exists(filename), f"{filename} exists"

            # Check the message is the correct rotation.
            message = f"message{self.__count-i-1}"
            with open(filename, "r") as stream:
                lines = stream.readlines()
                assert len(lines) == 1
                if message not in lines[0]:
                    logger.info(
                        f"filename {filename}\n  got {lines[0]}\n  but wanted {message}"
                    )
                assert message in lines[0]

        # assert not os.path.exists(f"{self.__logfile_directory}/logform.5")

    # --------------------------------------------------------------------------
    def configure_logging(self, settings=None):
        """
        Configure runtime logging, override base class.
        Presume that self._args is already set.
        """

        settings = {
            "console": {"enabled": True},
            "logfile": {"enabled": True},
            "max_bytes": 100,
            "mpqueue": {"enabled": True},
        }
        # Call the base method which has the extra kwarg.
        Mainiac.configure_logging(self, settings)

    # ----------------------------------------------------------
    def version(self):
        return "x.y.z"

    # ----------------------------------------------------------
    def about(self):
        return {"url": "/some/good/url", "description": "A good description"}

    # ----------------------------------------------------------
    def build_parser(self, arglist=None):

        # Make a parser.
        parser = argparse.ArgumentParser()

        # Return a structure describing the default subcommand.
        return parser



import argparse
import logging

import pytest

# Assertion helpers.
from dls_mainiac_lib.assert_helpers import (
    assert_parse_success,
    assert_parse_system_exit,
)

# Class under test.
from dls_mainiac_lib.mainiac import Mainiac

logger = logging.getLogger(__name__)


class Test_01_simple:

    # ----------------------------------------------------------------------------------------
    def test_01_simple(
        self,
        constants,
        logging_setup,
        output_directory,
    ):
        """
        Test mainiac base class.
        """

        failure_message = None
        try:

            # Instantiate the app class.
            app = _App()

            # Help.
            assert_parse_system_exit(
                "H",
                app,
                ["--help"],
                0,
                stdout_contains="positional arguments:",
            )

            # Version.
            assert_parse_system_exit(
                "V1", app, ["--version"], 0, stdout_contains="x.y.z"
            )
            assert_parse_system_exit(
                "V2", app, ["--version", "oops"], 0, stdout_contains="x.y.z"
            )
            assert_parse_system_exit(
                "V3", app, ["oops", "--version"], 0, stdout_contains="x.y.z"
            )

            # About.
            assert_parse_system_exit(
                "A1", app, ["--about"], 0, stdout_contains="/some/good/url"
            )
            assert_parse_system_exit(
                "A3",
                app,
                ["--about", "oops"],
                0,
                stdout_contains="/some/good/url",
            )
            assert_parse_system_exit(
                "A2",
                app,
                ["oops", "--about"],
                0,
                stdout_contains="/some/good/url",
            )

            # Unknown flags.
            e = "error: the following arguments are required: positional"
            assert_parse_system_exit(
                "U1",
                app,
                ["--unknown"],
                -1,
                stderr_contains=e,
            )

            # Missing required positional.
            e = "error: the following arguments are required: positional"
            assert_parse_system_exit(
                "U2",
                app,
                [],
                -1,
                stderr_contains=e,
            )

            # Too many positionals.
            assert_parse_system_exit(
                "T1",
                app,
                ["a", "b"],
                -1,
                stderr_contains="error: unrecognized arguments: b",
            )

            # Good command line.
            assert_parse_success("G1", app, ["positional1"])
            assert app._args.positional == "positional1"

        except Exception as exception:
            logger.exception(
                "unexpected exception during the test",
                exc_info=exception,
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)


# ---------------------------------------------------------------------------------
class _App(Mainiac):
    """
    App class.
    """

    def __init__(
        self,
    ):
        Mainiac.__init__(self, "test_01_simple")

    # ----------------------------------------------------------
    def version(self):
        """
        Method called from mainiac command line parsing.
        Should return string in form of N.N.N.
        """
        return "x.y.z"

    # ----------------------------------------------------------
    def about(self):
        """
        Method called from mainiac command line parsing.
        Should return dict which can be serialized by json.
        """

        return {"url": "/some/good/url", "description": "A good description"}

    # ----------------------------------------------------------
    def build_parser(self, arglist=None):
        """
        Method called from mainiac command line parsing.
        Should return argparser for this program.
        """

        # Make a parser.
        parser = argparse.ArgumentParser()

        # Add keyword arguments.
        parser.add_argument(
            "--keyword",
        )

        # Add positional arguments.
        parser.add_argument(
            type=str,
            dest="positional",
        )

        return parser



# Python standard logging.
import logging
import logging.handlers
import multiprocessing
import os
import threading

# Log formatter.
from dls_logformatter.dls_logformatter import DlsLogformatter


# -------------------------------------------------------------------------
class PermittedRotatingFileHandler(logging.handlers.RotatingFileHandler):
    """
    Override of logging class which creates logfiles with writable permissions.
    """

    def _open(self):
        # print("********** PermittedRotatingFileHandler constructor setting umask")
        # Save old umask.
        umask = os.umask(0)
        # Modify to allow all writable.
        os.umask(umask & ~0o0666)
        # Call base class method.
        rtv = logging.handlers.RotatingFileHandler._open(self)
        # Replace old umask.
        os.umask(umask)
        return rtv


# -------------------------------------------------------------------------
class Log666:
    # -------------------------------------------------------------------------
    def start_logfile(program_name, settings):
        """
        Start a rotating log in a standard logging directory location.
        """

        multiprocessing.current_process().name = program_name
        threading.current_thread().name = "main"

        # Place log messages in output directory named for this program.
        logfile_directory = "/tmp/logs/%s" % (program_name)

        if not os.path.exists(logfile_directory):
            # Make sure that created parent directories will have public permission.
            umask = os.umask(0)
            os.umask(umask & ~0o0777)
            os.makedirs(logfile_directory)
            os.umask(umask)

        logfile_filename = "%s/logform.log" % (logfile_directory)

        max_bytes = settings.get("max_bytes", 20000000)
        backup_count = settings.get("backup_count", 4)
        logfile_handler = PermittedRotatingFileHandler(
            logfile_filename, maxBytes=max_bytes, backupCount=backup_count
        )

        # Let logging write custom formatted messages to stdout.
        logfile_handler.setFormatter(DlsLogformatter())
        logging.getLogger().addHandler(logfile_handler)

        return logfile_handler



import argparse
import logging

import pytest

# Assertion helpers.
from dls_mainiac_lib.assert_helpers import (
    assert_parse_success,
    assert_parse_system_exit,
)

# Class under test.
from dls_mainiac_lib.mainiac import Mainiac

logger = logging.getLogger(__name__)


class Test_03_default_subcommand:

    # ----------------------------------------------------------------------------------------
    def test_03_default_subcommand(
        self,
        constants,
        logging_setup,
        output_directory,
        capsys,
    ):
        """
        Test mainiac base class.
        """

        failure_message = None
        try:

            # Instantiate the app class.
            app = _App()

            # Help.
            assert_parse_system_exit(
                "H",
                app,
                ["--help"],
                0,
                stdout_contains="positional arguments:",
            )

            # Version.
            assert_parse_system_exit(
                "V1", app, ["--version"], 0, stdout_contains="x.y.z"
            )
            assert_parse_system_exit(
                "V2", app, ["--version", "oops"], 0, stdout_contains="x.y.z"
            )
            assert_parse_system_exit(
                "V3", app, ["oops", "--version"], 0, stdout_contains="x.y.z"
            )

            # About.
            assert_parse_system_exit(
                "A1", app, ["--about"], 0, stdout_contains="/some/good/url"
            )
            assert_parse_system_exit(
                "A3",
                app,
                ["--about", "oops"],
                0,
                stdout_contains="/some/good/url",
            )
            assert_parse_system_exit(
                "A2",
                app,
                ["oops", "--about"],
                0,
                stdout_contains="/some/good/url",
            )

            # Unknown flags.
            e = "error: the following arguments are required: positional1"
            assert_parse_system_exit(
                "U1",
                app,
                ["--unknown"],
                -1,
                stderr_contains=e,
            )
            assert_parse_system_exit(
                "U2",
                app,
                ["--unknown", "x"],
                -1,
                stderr_contains="error: unrecognized arguments: --unknown",
            )

            # Subcommand with missing positional.
            e = "error: the following arguments are required: positional1"
            assert_parse_system_exit(
                "P1",
                app,
                ["subcommand1", "--keyword2=x"],
                -1,
                stderr_contains=e,
            )

            e = "error: the following arguments are required: positional1"
            assert_parse_system_exit(
                "P2",
                app,
                [],
                -1,
                stderr_contains=e,
            )

            # Subcommand with bad keyword.
            assert_parse_system_exit(
                "P3",
                app,
                ["subcommand1", "--unknown", "x"],
                -1,
                stderr_contains="error: unrecognized arguments: --unknown",
            )

            # Good command lines, verbose.
            case = "B1"
            assert_parse_success(case, app, ["--verbose", "subcommand1", "positional1"])
            assert app._args.verbose, case
            assert app._args.subcommand == "subcommand1", case
            assert app._args.keyword1 is None, case
            assert app._args.positional1 == "positional1", case

            case = "B2"
            assert_parse_success(case, app, ["subcommand1", "--verbose", "positional1"])
            assert app._args.verbose, case
            assert app._args.subcommand == "subcommand1", case
            assert app._args.keyword1 is None, case
            assert app._args.positional1 == "positional1", case

            case = "B3"
            assert_parse_success(case, app, ["subcommand1", "positional1", "--verbose"])
            assert app._args.verbose, case
            assert app._args.subcommand == "subcommand1", case
            assert app._args.keyword1 is None, case
            assert app._args.positional1 == "positional1", case

            # Good command lines.
            case = "G1"
            assert_parse_success(case, app, ["subcommand1", "positional1"])
            assert app._args.subcommand == "subcommand1", case
            assert app._args.keyword1 is None, case
            assert app._args.positional1 == "positional1", case

            case = "G2"
            assert_parse_success(
                case, app, ["subcommand1", "--keyword1=x", "--verbose", "positional1"]
            )
            assert app._args.verbose, case
            assert app._args.subcommand == "subcommand1", case
            assert app._args.keyword1 == "x", case
            assert app._args.positional1 == "positional1", case

            case = "G3"
            assert_parse_success(case, app, ["subcommand2", "positional2"])
            assert app._args.subcommand == "subcommand2", case
            assert app._args.keyword2 is None, case
            assert app._args.positional2 == "positional2", case

            assert_parse_success(
                "G4", app, ["subcommand2", "--keyword2=y", "positional2"]
            )

            # Default subcommand.
            case = "G5"
            assert_parse_success(case, app, ["positional1"])
            assert app._args.subcommand == "subcommand1", case
            assert app._args.keyword1 is None, case
            assert app._args.positional1 == "positional1", case

            case = "G6"
            assert_parse_success(case, app, ["--keyword1=x", "positional1"])
            assert app._args.subcommand == "subcommand1", case
            assert app._args.keyword1 == "x", case
            assert app._args.positional1 == "positional1", case

        except Exception as exception:
            logger.exception("unexpected exception during the test", exc_info=exception)
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)


# ---------------------------------------------------------------------------------
class _App(Mainiac):
    """
    App class.
    """

    def __init__(
        self,
    ):
        Mainiac.__init__(self, "test_02_subcommand")

    # ----------------------------------------------------------
    def version(self):
        return "x.y.z"

    # ----------------------------------------------------------
    def about(self):
        return {"url": "/some/good/url", "description": "A good description"}

    # ----------------------------------------------------------
    def build_parser(self, arglist=None):

        # Make a parser.
        parser = argparse.ArgumentParser()

        subparsers = parser.add_subparsers(help="subcommands", dest="subcommand")
        subparsers.required = True

        subparser1 = subparsers.add_parser("subcommand1", help="Subcommand 1.")
        # Add keyword arguments.
        subparser1.add_argument(
            "--keyword1",
        )

        # Add positional arguments.
        subparser1.add_argument(
            type=str,
            dest="positional1",
        )

        subparser2 = subparsers.add_parser("subcommand2", help="Subcommand 2.")
        # Add keyword arguments.
        subparser2.add_argument(
            "--keyword2",
        )

        # Add positional arguments.
        subparser2.add_argument(
            type=str,
            dest="positional2",
        )

        # Return a structure describing the default subcommand.
        return {
            "parser": parser,
            "subcommand_dest": "subcommand",
            "default_subcommand": "subcommand1",
        }
